// sxanth@ceid.upatras.gr
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ncurses.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>

#include "dbstree.h"
#include "inttree.h"

// If you don't want an external viewer,
// comment the following line
#define SOURCE_VIEWER "indent -st -kr | less"
//#define SOURCE_VIEWER "less"

// "uti.mod"
//
// utilities
//

static char *strDup (char *s)
{
	return strcpy (new char [strlen (s) + 1], s);
}

#define SHUT_UP_GCC(x) (void)x;

static void progress ()
{
static	char pchars [] = "-\\|/", out [] = { '\r', '[', 0, ']' };
static	int pi, ti;
	if (!(ti++ & 127)) {
		out [2] = pchars [pi = (pi + 1)%(sizeof pchars - 1)];
		write (fileno (stdout), out, sizeof out);
	}
}

class foreach_intNode
{
   protected:
virtual	void A (intNode*) = 0;
	void inorder (intNode*);
   public:
};

void foreach_intNode::inorder (intNode *i)
{
	if (i->less) inorder (i->less);
	A (i);
	if (i->more) inorder (i->more);
}


class load_file
{
	int fd;
   public:
	load_file (char*);
	int success;
	char *data;
	int len;
	~load_file ();
};

#define ZC_OK 0
#define ZC_NA 1
#define ZC_AC 2
#define ZC_FF 3

load_file::load_file (char *f)
{
	data = NULL;
	success = ZC_NA;
	fd = -1;

	struct stat statbuf;
	if (stat (f, &statbuf) == -1) return;
	len = statbuf.st_size;
	if (len == -1 || (fd = open (f, O_RDONLY)) == -1) return;

	success = ((data = (char*) mmap (0, len, PROT_READ, MAP_PRIVATE, fd, 0))
		 != MAP_FAILED)	? ZC_OK : ZC_FF;
}

load_file::~load_file ()
{
	if (data && data != MAP_FAILED) munmap (data, len);
	if (fd != -1) close (fd);
}
// "st1.mod"
//
// Symbol tree
//

class symNode : public dbsNodeStr
{
   public:
	symNode *less, *more;
	unsigned int i;
	symNode ();
	~symNode ();
};

static dbsTree<symNode> symTree;
static unsigned int nsym = 1;

symNode::symNode () : dbsNodeStr ()
{
	symTree.addself (this);
	i = nsym++;
}

symNode::~symNode ()
{
	if (less) delete less;
	if (more) delete more;
}

unsigned int enter_symbol (char *sy)
{
	DBS_STRQUERY = sy;
	symNode *s = (symNode*) symTree.dbsFind ();
	if (!s) {
		DBS_STRQUERY = strDup (DBS_STRQUERY);
		s = new symNode;
		progress ();
	}
	return s->i;
}
//
// Hyper nodes
//

class hyperNode : public intNode
{
   public:
	void settle ();
	intTree itree;
	unsigned int *rel;
	hyperNode ();
	void relates (unsigned int);
	void rsettle ();
};

static intTree hyperTree;

hyperNode::hyperNode () : intNode (&hyperTree)
{ }

#define SOURCE_ID	0
#define STRUCT_ID	1
#define CALLER_ID	2
#define GLOBAL_ID	3
#define CALL_ID		4
#define MEMBER_ID	5
#define NFIELDS		6
#define RECURS_BOOST	7	/* color */

#define BASE		(1024*1024)
#define SOURCE_BASE	(SOURCE_ID * BASE)
#define CALLER_BASE	(CALLER_ID * BASE)
#define GLOBAL_BASE	(GLOBAL_ID * BASE)
#define CALL_BASE	(CALL_ID * BASE)
#define STRUCT_BASE	(STRUCT_ID * BASE)
#define MEMBER_BASE	(MEMBER_ID * BASE)
#define FIELD_BASE	(NFIELDS * BASE)
#define RECURS_BASE	(RECURS_BOOST * BASE)

static unsigned int symbol_of (unsigned int i)
{
	return i % BASE;
}

static unsigned int base_of (unsigned int i)
{
	return i / BASE;
}

unsigned int counts [NFIELDS];

void hyperNode::relates (unsigned int y)
{
	if (!itree.intFind (y)) new intNode (&itree);
}

static void relate (int x, int y)
{
	hyperNode *h = (hyperNode*) hyperTree.intFind (x);

	if (!h) {
		h = new hyperNode;
		counts [base_of (x)]++;
	}
	h->relates (y);
}

static int nlinks = 0;

static void relative (int x, int y)
{
	++nlinks;
	relate (base_of (x) == CALLER_ID ? x - CALLER_BASE + CALL_BASE : x, y);
	relate (base_of (y) == CALLER_ID ? y - CALLER_BASE + CALL_BASE : y, x);
}
// "ln.mod"
//
// Line numbers
//

struct file_chunk {
	unsigned int start, end;
} NoChunk = { ~0, ~0 };

class lnnode : public intNode
{
   public:
	file_chunk line;
	lnnode (unsigned int, unsigned int);
};

static intTree lnTree;

lnnode::lnnode (unsigned int l, unsigned int e) : intNode (&lnTree)
{
	line.start = l;
	line.end = e;
}

static void enter_line (unsigned int f, unsigned int l, unsigned int e)
{
	if (!lnTree.intFind (f)) new lnnode (l, e);
}

file_chunk line_of (unsigned int f)
{
	lnnode *l = (lnnode*) lnTree.intFind (f);
	return (l) ? l->line : NoChunk;
}
// "agr.mod"
//
// Make structure -> members link
//

void struct_fp_link (char *s)
{
	char *dot, ss [100];

	if (!(dot = strchr (s, '.'))) return;
	*dot = 0; strcpy (ss, s+1); *dot = '.';
	relative (CALL_BASE + enter_symbol (s),
		  STRUCT_BASE + enter_symbol (ss));
}

void struct_link (char *s)
{
	char *dot, ss [100];

	if (!(dot = strchr (s, '.'))) return;
	*dot = 0; strcpy (ss, s); *dot = '.';
	relative (MEMBER_BASE + enter_symbol (s),
		  STRUCT_BASE + enter_symbol (ss));
}
//
// Symbol Table
//

static struct st_entry {
	char *name;
	int weight;
} *symbol_table;

static void totable (symNode *s)
{
static	int order = 0;
	symbol_table [s->i].name = s->Name;
	symbol_table [s->i].weight = order++;
}

void make_symbol_table ()
{
	symbol_table = new st_entry [nsym];
	symTree.foreach (totable);
	delete symTree.root;
}

static int weight (unsigned int i)
{
	return symbol_table [symbol_of (i)].weight;
}

static char *symbol_name (unsigned int i)
{
	return symbol_table [symbol_of (i)].name;
}
// "hns.mod"
//
// Settle hypernodes
//

static unsigned int *tmp_table;

class enter_intnodes : public foreach_intNode
{
	void A (intNode *i) {
		*tmp_table++ = i->Key;
	}
   public:
	enter_intnodes (intTree *I) { if (I->root) inorder (I->root); }
};

void hyperNode::settle ()
{
	if (!itree.root)
		return;
	tmp_table = rel = new unsigned int [itree.cnt];
	enter_intnodes E (&itree);
	SHUT_UP_GCC (E)
	delete itree.root;
}

class hyper_settle : public foreach_intNode
{
   protected:
	void A (intNode *i) {
		((hyperNode*)i)->settle ();
	}
   public:
	hyper_settle () { inorder (hyperTree.root); }
};

void settle_hyperNodes ()
{
	hyper_settle H;
	SHUT_UP_GCC (H)
}
//
// Sort symbol indexes
//

static void swap (unsigned int v[], int i, int j)
{
	unsigned int swp = v [i];
	v [i] = v [j];
	v [j] = swp;
}

void qsort (unsigned int v[], int left, int right)
{
	int i, last, wleft;

	if (left >= right)
		return;
	swap (v, left, (left+right)/2);
	wleft = weight (v [left]);
	last = left;
	for (i = left+1; i <= right; i++)
		if (weight (v [i]) < wleft)
			swap (v, i, ++last);
	swap (v, last, left);
	qsort (v, left, last-1);
	qsort (v, last+1, right);
}

void sort_fields (unsigned int v [], int n)
{
	int i, j, k;
	struct {
		unsigned int *v;
		int n;
	} field [NFIELDS];

	for (i = 0; i < NFIELDS; i++) {
		field [i].v = (unsigned int*) alloca (n * sizeof (int));
		field [i].n = 0;
	}

	for (i = 0; i < n; i++)
		field [base_of (v [i])].v [field [base_of (v [i])].n++] = v [i];

	for (j = i = 0; i < NFIELDS; i++) {
		if (field [i].n > 1)
			qsort (field [i].v, 0, field [i].n - 1);
		for (k = 0; k < field [i].n; k++)
			v [j++] = field [i].v [k];
	}
}
// "ia.mod"
//
// Staring tables
//

class hyper_gets : public foreach_intNode
{
	void A (intNode*);
   public:
	unsigned int nfiles, nfuncs;
	hyper_gets ();
};

unsigned int *Files, *funcs, *entries, nentries;

static bool has_callers (hyperNode *h)
{
	int i;
	for (i = 0; i < h->itree.cnt; i++)
		if (base_of (h->rel [i]) == CALLER_ID
		// recursive callers of self, do not count
		&& symbol_of (h->rel [i]) != symbol_of (h->Key))
			return true;
	return false;
}

void hyper_gets::A (intNode *i)
{
	hyperNode *h = (hyperNode*) i;

	switch (base_of (h->Key)) {
		case SOURCE_ID: Files [nfiles++] = h->Key; break;
		case CALL_ID: funcs [nfuncs++] = h->Key;
			if (!has_callers (h)) nentries++; break;
	}
}

hyper_gets::hyper_gets ()
{
	nfiles = nfuncs = nentries = 0;
	inorder (hyperTree.root);
}

void starting_tables ()
{
	unsigned int i, j;
	Files = new unsigned int [counts [SOURCE_ID]];
	funcs = new unsigned int [counts [CALL_ID]];
	hyper_gets H;
	SHUT_UP_GCC (H);
	entries = new unsigned int [nentries];
	for (i = j = 0; i < counts [CALL_ID]; i++)
		if (!has_callers ((hyperNode*) hyperTree.intFind (funcs [i])))
			entries [j++] = funcs [i];
	qsort (Files, 0, counts [SOURCE_ID] - 1);
	qsort (funcs, 0, counts [CALL_ID] - 1);
	qsort (entries, 0, nentries - 1);
}
// "pf.mod"
//
// Parse File
//

void linenumber (char *s)
{
	char *f;
	if (!(f = strchr (s, ' ')))
		return;
	*f = 0;
	enter_line (enter_symbol (s), atoi (f + 1), atoi (strchr (f+1, ' ')));
}

#define FMT_CALLER	'D'
#define FMT_CALL	'F'
#define FMT_SOURCE	'P'
#define FMT_GLOBAL	'G'
#define FMT_STRUCT	'S'
#define FMT_LINE	'L'

unsigned int parse_line (char *l)
{
	int base;

	switch (l [0]) {
		case FMT_LINE:	 base = CALL_BASE;
				 linenumber (l+3);	break;
		case FMT_CALLER: base = CALLER_BASE;	break;
		case FMT_CALL:   base = CALL_BASE;
				 if (l[3] == '*')
				 struct_fp_link (l+3);	break;
		case FMT_SOURCE: base = SOURCE_BASE;	break;
		case FMT_GLOBAL: base = GLOBAL_BASE;	break;
		case FMT_STRUCT: base = MEMBER_BASE;
				 struct_link (l+3);	break;
		default: return ~0;
	}

	return base + enter_symbol (l + 3);
}

#define	MLINE 512
void parse_file (FILE *f)
{
static	int dfunc = 0;
	int h;
	char line [MLINE];


	while (fgets (line, MLINE, f)) {
		line [strlen (line) - 1] = 0;
		if ((h = parse_line (line)) == ~0) continue;
		if (base_of (h) == CALLER_ID || base_of (h) == SOURCE_ID)
			dfunc = h;
		else relative (dfunc, h);
	}
}

void read_file (char *p)
{
	FILE *f = fopen (p, "r");

	if (!f) {
		printf ("No such file [%s]\n", p);
		exit (1);
	}
	parse_file (f);
	fclose (f);
	if (!hyperTree.root) {
		printf ("File [%s] empty\n", p);
		exit (0);
	}
	make_symbol_table ();
	settle_hyperNodes ();
	starting_tables ();
}
//
// Get Relations of Link
//

struct linker {
	unsigned int tn, *t;
	linker (unsigned int);
};

linker::linker (unsigned int i)
{
	hyperNode *h = (hyperNode*) hyperTree.intFind (i);
	sort_fields (t = h->rel, tn = h->itree.cnt);
}
// "recur.mod"
//
// Recursion detection
//

class recursion_detector
{
	unsigned int caller;
	bool check_in (unsigned int);
   public:
	recursion_detector (unsigned int, unsigned int);
	unsigned int *visited;
	bool indeed;
	~recursion_detector ();
};

bool recursion_detector::check_in (unsigned int c)
{
	hyperNode *h = (hyperNode*) hyperTree.intFind (c);
	unsigned int cs = symbol_of (c);
	unsigned int *t = h->rel;
	unsigned int tn = h->itree.cnt;
	unsigned int i;

	for (i = 0; i < tn; i++)
		if (base_of (t [i]) == CALL_ID) {
			unsigned int s = symbol_of (t [i]);
			if (s == caller) {
				visited [caller] = cs;
				return true;
			} else if (!visited [s]) {
				visited [s] = cs;
				if (check_in (t [i])) return true;
			}
		}
	return false;
}

recursion_detector::recursion_detector (unsigned int ca, unsigned ch)
{
	visited = (unsigned int*) calloc (nsym, sizeof (int));
	caller = symbol_of (ca);
	visited [symbol_of (ch)] = caller;

	indeed = check_in (ch);
}

recursion_detector::~recursion_detector ()
{
	free (visited);
}

bool is_recursive (unsigned int caller, unsigned int child)
{
	recursion_detector RD (caller, child);
	return RD.indeed;
}
//
// make recursion path
//

struct recursion_path
{
	unsigned int *path;
	unsigned int n;
	recursion_path (unsigned int, unsigned int);
	~recursion_path ();
};

recursion_path::recursion_path (unsigned int caller, unsigned int child)
{
	recursion_detector RD (caller, child);

	if (RD.indeed) {
		unsigned int cs = symbol_of (caller);
		unsigned int i, cnt;

		for (cnt = 2, i = RD.visited [cs]; i != cs; i = RD.visited [i])
			++cnt;
		n = cnt;
		path = (unsigned int*) malloc (sizeof (int) * cnt);
		path [n-1] = cs + CALL_BASE;
		for (cnt = 1, i = RD.visited [cs]; i != cs; i = RD.visited [i])
			path [n-1-cnt++] = i + CALL_BASE;
		path [n-1-cnt++] = i + CALLER_BASE;
		n = cnt;
	} else path = 0;
}

recursion_path::~recursion_path ()
{
	if (path) free (path);
}
// "grf.mod"
//
// Graphics
//

unsigned int scr_x, scr_y;

enum APP_COLOR {
	BLACK, HIGHLIGHT, C_UP1, C_UP2, C_UP3, NORMAL, C_DOWN1,
	C_DOWN2, C_DOWN3, C_DOWN4, OTHER
};

enum ecol {
	WHITE, BLUE, RED, GREEN, YELLOW, CYAN, MAGENTA
};

int Gcolor (APP_COLOR a)
{
	switch (a) {
	case HIGHLIGHT:	return COLOR_PAIR(WHITE)|A_BOLD;
	case C_UP1:	return COLOR_PAIR (BLUE);
	case C_UP2:	return COLOR_PAIR (BLUE)|A_BOLD;
	case C_UP3:	return COLOR_PAIR (RED);
	case NORMAL:	return COLOR_PAIR(WHITE);
	case C_DOWN1:	return COLOR_PAIR(YELLOW)|A_DIM;
	case C_DOWN2:	return COLOR_PAIR(WHITE)|A_DIM;
	case C_DOWN3:	return COLOR_PAIR (CYAN);
	case C_DOWN4:	return COLOR_PAIR(WHITE)|A_DIM;
	default:
	case OTHER:	return COLOR_PAIR(MAGENTA);
	}
}

void printstr (int x, int y, char *s, APP_COLOR a)
{
	attrset (Gcolor (a));
	move (y, x);
	/* unfortunatelly, there's a bug in ncurses and
	   addnstr (s, -1); counts tabs as 1 character
	   with the result that long lines go past the end
	   of the scr_x and appear on the next line
	 */
	unsigned int i;
	char *p;
	for (i = x, p = s; i < scr_x && *p; p++)
		if (*p == '\t') i += 8;
		else i++;
	if (!*p) printw ("%s", s);
	else {
		char tmp [256];
		strncpy (tmp, s, p - s);
		tmp [p - s] = 0;
		printw ("%s", tmp);
	}
}

void cls ()
{
	clear ();
}

int inkey ()
{
	return getch ();
}

// Init modes

void init_ncurses ()
{
	initscr ();
	//leaveok (stdscr, true);
	start_color ();
	init_pair (BLUE, COLOR_BLUE, COLOR_BLACK);
	init_pair (RED, COLOR_RED, COLOR_BLACK);
	init_pair (GREEN, COLOR_GREEN, COLOR_BLACK);
	init_pair (YELLOW, COLOR_YELLOW, COLOR_BLACK);
	init_pair (CYAN, COLOR_CYAN, COLOR_BLACK);
	init_pair (MAGENTA, COLOR_MAGENTA, COLOR_BLACK);
	crmode ();
	keypad (stdscr, TRUE);
	noecho ();
	scr_x = COLS;
	scr_y = LINES-1;
}

void init_graphics ()
{
	init_ncurses ();
}

void end_graphics ()
{
	flash ();
	endwin ();
}
//
// Text entry display
//

static bool has_calls (unsigned int i)
{
	hyperNode *h = (hyperNode*) hyperTree.intFind (i);
	unsigned int *a = h->rel, j = h->itree.cnt;

	for (i = 0; i < j; i++)
		if (base_of (a [i]) == CALL_ID) return true;
	return false;
}

static bool has_file (unsigned int i)
{
	hyperNode *h = (hyperNode*) hyperTree.intFind (i);
	unsigned int *a = h->rel, j = h->itree.cnt;

	for (i = 0; i < j; i++)
		if (base_of (a [i]) == SOURCE_ID) return true;
	return false;
}

APP_COLOR colors (unsigned int v)
{
	switch (base_of (v)) {
	case SOURCE_ID: return C_UP1;
	case CALLER_ID: return C_UP2;
	case MEMBER_ID: return C_DOWN4;
	case GLOBAL_ID: return C_UP3;
	case NFIELDS:	return OTHER;
	case CALL_ID:   return !has_file (v) ? C_DOWN3 :
			       has_calls (v) ? C_DOWN1 : C_DOWN2;
	case RECURS_BOOST: return C_UP3;
	}
	return OTHER;
}

void printent (int x, int y, unsigned int e, unsigned int m, bool hi)
{
	char trunc [80], *dot;
	char *s = symbol_name (e);

	if (strlen (s) > m)
		s = strcpy (trunc, s + strlen (s) - m);
	printstr (x, y, s, hi ? HIGHLIGHT : colors (e));
	if ((dot = strchr (s, '.')))
		printstr (x + (dot - s), y, ".", NORMAL);
}
// "tdm.mod"
//
// Map with ncurses
//

struct coords {
	unsigned int x, y;
};

class TDmap {
	unsigned int epl, eps, e_spc;
	unsigned int *vmap, e_tot, e_cur, e_top, e_topp;
	coords locate (unsigned int);
   public:
	TDmap (unsigned int*, unsigned int, unsigned int, unsigned int=0);
	void move_arrow (int);
	void move_jump (unsigned int);
	void draw_map ();
	unsigned int enter ();
	unsigned int cursor ();
};

unsigned int TDmap::enter ()
{
	if (base_of (vmap [e_cur]) != RECURS_BOOST)
		return vmap [e_cur];
	return symbol_of (vmap [e_cur]) + CALL_BASE;
}

unsigned int TDmap::cursor ()
{
	return e_cur;
}

void TDmap::move_arrow (int a)
{
	unsigned int dir = 0, mov;
	switch (a) {
		case KEY_LEFT:	mov = 1, dir = 1; break;
		case KEY_RIGHT:	mov = 1; break;
		case KEY_UP:	mov = epl, dir = 1; break;
		case KEY_DOWN:	mov = epl; break;
		case KEY_PPAGE:	mov = eps, dir = 1; break;
		case KEY_NPAGE: mov = eps; break;
		default: 	return;
	}

	if (!dir) {
		if (e_cur+mov < e_tot) e_cur += mov;
		else if (mov == eps)
			while (e_cur + epl < e_tot) e_cur += epl;
		if (e_cur >= e_top + eps) e_top += eps;
	} else {
		e_cur = (e_cur >= mov) ? e_cur - mov : e_cur % epl;
		if (e_cur < e_top) e_top = (e_top > eps) ? e_top - eps : 0;
	}
}

void TDmap::move_jump (unsigned int i)
{
	e_cur = i;
	e_top = (i/eps) * eps;
}

coords TDmap::locate (unsigned int i)
{
	coords r;
	r.x = e_spc * (i % epl);
	r.y = (i - e_top) / epl;
	return r;
}

void TDmap::draw_map ()
{
	unsigned int i;
	coords xy;

	if (e_topp != e_top) {
		e_topp = e_top;
		cls ();
	}
	for (i = e_top; i < e_top+eps && i < e_tot; i++) {
		xy = locate (i);
		printent (xy.x, xy.y, vmap [i], e_spc-1, false);
	}
	xy = locate (e_cur);
	printent (xy.x, xy.y, vmap [e_cur], e_spc-1, true);
}

unsigned int fieldlen (unsigned int *v, unsigned int n)
{
	unsigned int i, ml;

	for (ml = i = 0; i < n; i++)
		if (strlen (symbol_name (v [i])) > ml)
			ml = strlen (symbol_name (v [i]));

	return (ml > scr_x/2) ? scr_x/2 : (ml < 15) ? 15 : ml+1;
}

TDmap::TDmap (unsigned int *v, unsigned int n, unsigned int c, unsigned int ml)
{
	if (!ml) ml = fieldlen (v, n);
	e_spc = scr_x/(scr_x/ml);
	vmap = v;
	e_topp = e_tot = n;
	epl = scr_x / e_spc;
	eps = scr_y * epl;
	move_jump (e_cur = c);
}
// "tv.mod"
//
// text Viewer
//

#ifndef SOURCE_VIEWER

class txtviewer
{
	char **line;
	unsigned int nlines, maxline, topline, tab;
	void draw ();
	void scrolll (int);
	void main ();
   public:
	txtviewer (char*, char*, unsigned int);
};

void txtviewer::draw ()
{
	unsigned int i;
	cls ();
	for (i = topline; i < nlines && i < topline+scr_y; i++) {
		if (strlen (line [i]) <= tab) continue;
		printstr (0, i-topline, line [i]+tab, NORMAL);
	}
}

void txtviewer::scrolll (int key)
{
	unsigned int tlp = topline, tbp = tab;
	switch (key) {
	case KEY_UP:
		if (topline > 0) topline--;
		break;
	case KEY_DOWN:
		if (topline + scr_y < nlines) topline++;
		break;
	case KEY_LEFT:
		if (tab) tab--;
		break;
	case KEY_RIGHT:
		if (maxline - tab > scr_x/2) tab++;
		break;
	case KEY_PPAGE:
		if (topline) topline = (topline > scr_y) ? topline - scr_y : 0;
		break;
	case KEY_NPAGE:
		if (topline + scr_y + 1 < nlines)
			topline += scr_y;
		break;
	default:
		return;
	}
	if (tlp != topline || tbp != tab) draw ();
}

void txtviewer::main ()
{
	int key;
	while ((key = inkey ()) != '\n' && key != ' ')
		scrolll (key);
}

txtviewer::txtviewer (char *title, char *txt, unsigned int l)
{
	nlines = l+1;
	line = (char**) alloca (nlines * sizeof (char*));
	maxline = l = 0;
	line [l++] = title;
	do {
		line [l] = txt;
		if ((txt = strchr (txt, '\n')))
			*txt++ = 0;
		if (strlen (line [l]) > maxline)
			maxline = strlen (line [l]);
	} while (++l < nlines);
	topline = tab = 0;

	draw ();
	main ();
}

#else		/* SOURCE_VIEWER exists */

void source_viewer_fatal ()
{
	endwin ();
	printf ("ALERT! Failed executing ["SOURCE_VIEWER"] to display the text.\n"
		" Please recompile nccnav with a different SOURCE_VIEWER value.\n");
	exit (1);
}

void txtviewer (char *title, char *txt, unsigned int l)
{
	FILE *p = popen (SOURCE_VIEWER, "w");
	if (!p) source_viewer_fatal ();
	fputs (title, p); fputc ('\n', p);
	fputs (txt, p);
	fflush (p);
	pclose (p);
}

#endif
// "vf.mod"
//
// Extract text from file
//

char *text_from (char *f, int l1, int l2)
{
	load_file L (f);

	if (L.success != ZC_OK) return NULL;

	int i, cl;
	char *r;

	for (i = 0, cl = 1; i <= L.len && cl < l1; i++)
		if (L.data [i] == '\n') cl++;
	if (cl < l1) return NULL;
	l1 = i;
	for (;i < L.len && cl <= l2; i++)
		if (L.data [i] == '\n') cl++;
	if (cl < l2) return NULL;
	l2 = i;

	r = new char [1 + l2 - l1];
	memcpy (r, L.data + l1, l2 - l1);
	r [l2 - l1] = 0;
	return r;
}

void func_text (unsigned int f)
{
	char *txt, *title, *fn;
	file_chunk F = line_of (symbol_of (f));

	if (F.start == ~0U) return;
	linker L (f);
	fn = symbol_name (L.t [0]);
	txt = text_from (fn, F.start, F.end);
	if (txt) {
		title = (char*) alloca (strlen (fn) + 20);
		sprintf (title, "# %s %u", fn, F.start);
		txtviewer (title, txt, F.end - F.start + 1);
		delete [] txt;
	}
}
// "file"
//
// view entire file
//

void file_text (unsigned int i)
{
#ifdef SOURCE_VIEWER
	load_file L (symbol_name (i));

	if (L.success != ZC_OK) return;

	FILE *p = popen (SOURCE_VIEWER, "w");
	if (!p) source_viewer_fatal ();
	fwrite (L.data, 1, L.len, p);
	fflush (p);
	pclose (p);
#endif
}
// "hi.mod"
//
// History
//

static struct history
{
	history *prev;
	unsigned int i;
	history (unsigned int);
	~history ();
} *history_top;

int nhist, history_request;

history::history (unsigned int e)
{
	i = e;
	prev = history_top;
	history_top = this;
	nhist++;
	history_request++;
}

history::~history ()
{
	history_top = prev;
	nhist--;
}
// "mod.mod"
//
// modes
//

void historymode ()
{
	unsigned int *a = (unsigned int*) alloca (nhist * sizeof (int));
	int i;
	history *hp;

	for (hp = history_top, i = 0; i < nhist; i++, hp = hp->prev)
		a [nhist - i - 1] = hp->i;

	TDmap M (a, nhist, nhist - 1, scr_x);

	M.draw_map ();
	for (;; M.draw_map ()) {
		if ((i = inkey ()) == '\n') {
			history_request = M.cursor ()+1;
			return;
		}
		if (i == 'q') break;
		if (i == ' ') {
			unsigned int b = base_of (M.enter ());
			if (b == CALL_ID) func_text (M.enter ());
			else if (b == CALLER_ID)
				func_text (symbol_of (M.enter ()) + CALL_BASE);
			cls ();
		}
		else M.move_arrow (i);
	}
}

void othermode (unsigned int);

void recursmode (recursion_path &RP)
{
	int i;
	TDmap M (RP.path, RP.n, 0, scr_x);

	M.draw_map ();
	for (; nhist <= history_request; M.draw_map ()) {
		printstr (0, scr_y, "Recursion unroll", C_UP2);
		if ((i = inkey ()) == 'q')
			break;
		if (i == '\n') {
			if (M.cursor () == 0) break;
			othermode (M.enter ());
			cls ();
		} else M.move_arrow (i);
	}
}

void color_recursives (unsigned int caller, unsigned int *pa, int npa)
{
	int i;
	for (i = 0; i < npa; i++)
		if (base_of (pa [i]) == CALL_ID && is_recursive (caller, pa [i]))
			pa [i] = symbol_of (pa [i]) + RECURS_BASE;
}

class FOO {};

void othermode (unsigned int e)
{
static	bool rdetect = true;

	int key;
	unsigned int i, j, c, *jarr;
	history H (e);

	e = base_of (e) == CALLER_ID ? symbol_of (e) + CALL_BASE : e;
	bool cisf = base_of (e) == CALL_ID;
	bool cisp = base_of (e) == SOURCE_ID;
	linker L (e);

	jarr = (unsigned int*) alloca ((L.tn + 1) * sizeof(int));
	i = j = 0;
	while (base_of (L.t [i]) < CALL_ID && i < L.tn)
		jarr [j++] = L.t [i++];
	jarr [c = j++] = e = symbol_of (e) + FIELD_BASE;
	while (i < L.tn)
		jarr [j++] = L.t [i++];

	if (rdetect && cisf)
		color_recursives (e, jarr + c + 1, i - c);

	TDmap M (jarr, j, c);

	M.draw_map ();
	for (;nhist <= history_request; M.draw_map ()) {
		if (cisf && c >= 2 && base_of (jarr [1]) == SOURCE_ID) {
			printstr (0, scr_y, "* FUNCTION IS       :MULTIPLE DEFINITIONS *", HIGHLIGHT);
			printstr (13, scr_y,             " br0ken", C_UP2);
		}
		switch (key = inkey ()) {
		case ' ':
			j = base_of (M.enter ());
			if (j == CALL_ID) func_text (M.enter ());
			else if (j == CALLER_ID)
				func_text (symbol_of (M.enter ()) + CALL_BASE);
			else if (j == NFIELDS && cisf)
				func_text (symbol_of (M.enter ()) + CALL_BASE);
			else if (j == SOURCE_ID || (j == NFIELDS && cisp))
				file_text (symbol_of (M.enter ()));
			else break;
			cls ();
			break;
		case '\n':
			if (M.enter () == e) return;
			othermode (M.enter ());
			cls ();
			break;
		case KEY_BACKSPACE:
			cls ();
			throw FOO();
		case 'q':
			cls ();
			return;
		case 'h':
			historymode ();
			cls ();
			break;
		case 'r':
			if (base_of (M.enter ()) == CALL_ID) {
				recursion_path RP (e, M.enter ());
				if (RP.path) recursmode (RP);
			}
			cls ();
			break;
		case 'R':
			if ((rdetect =! rdetect)) {
				printstr (0, scr_y, "Recursion Detector ENABLED.", OTHER);
				color_recursives (e, jarr + c + 1, i - c);
			} else {
				printstr (0, scr_y, "Recursion Detector DISABLED", C_UP3);
				for (j = c; j < i+1; j++)
					if (base_of (jarr [j]) == RECURS_BOOST)
						jarr [j] -= RECURS_BASE - CALL_BASE;
			}
			break;
		default:
			M.move_arrow (key);
		}
	}
}

void enter_othermode (unsigned int e)
{
	try {
		othermode (e);
	} catch (FOO) { }
}

void funcmode (bool all)
{
	unsigned int *t, tn;
	if (all) t = funcs, tn = counts [CALL_ID];
	else t = entries, tn = nentries;

	if (!tn) {
		flash ();
		return;
	}

	int key;
	TDmap M (t, tn, 0);

	M.draw_map ();
	for (;;M.draw_map ()) switch (key = inkey ()) {
	case '\n':
		enter_othermode (M.enter ());
		cls ();
		break;
	case KEY_BACKSPACE:
	case 'q':
		return;
	case ' ':
		func_text (M.enter ());
		cls ();
		break;
	default:
		if (isalpha (key)) {
			unsigned int j = M.cursor ();
			if (*symbol_name (t [j]) < key)
				while (j < tn
				&& *symbol_name (t [j]) < key)
					j++;
			else
				while (j > 0
				&& *symbol_name (t [j]) > key)
					j--;
			M.move_jump (j);
		} else M.move_arrow (key);
	}
}

void initialmode ()
{
	int key;
	TDmap M (Files, counts [SOURCE_ID], 0);

	M.draw_map ();
	for (;;M.draw_map ()) switch (key = inkey ()) {
		case '\n':
			enter_othermode (M.enter ());
			cls ();
			break;
		case KEY_BACKSPACE:
		case 'q':
			return;
		case 'E':
			funcmode (false);
			cls ();
			break;
		case 'O':
			funcmode (true);
			cls ();
			break;
		default:
			M.move_arrow (key);
	}
}
// "main"
//

int main (int argc, char **argv)
{
	read_file ((argc > 1) ? argv [1] : (char*)"Code.map");
	init_graphics ();
	initialmode ();
	end_graphics ();
	printf ("%i\tFilez \n%i\tfunctionz \n%i\tglobal variables \n%i\tMembers \n%i\tTotal links\n",
		counts [SOURCE_ID], counts [CALL_ID], counts [GLOBAL_ID],
		counts [MEMBER_ID], nlinks);
}
